/* Wayland compositor running on top of an X server.

Copyright (C) 2022 to various contributors.

This file is part of 12to11.

12to11 is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

12to11 is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with 12to11.  If not, see <https://www.gnu.org/licenses/>.  */

#include <stdio.h>
#include <string.h>

#include "compositor.h"

typedef struct _DataSource DataSource;
typedef struct _DataDeviceReference DataDeviceReference;
typedef struct _DataOffer DataOffer;

struct _DataOffer
{
  /* The next data offer object in this list.  */
  DataOffer *next, *last;

  /* The DataSource corresponding to this data offer.  */
  DataSource *source;

  /* The struct wl_resource of this data offer.  */
  struct wl_resource *resource;
};

struct _DataDeviceReference
{
  /* The next and last data device references.  */
  DataDeviceReference *next, *last;

  /* The associated data device.  */
  DataDevice *device;

  /* The associated struct wl_resource.  */
  struct wl_resource *resource;
};

struct _DataSource
{
  /* List of const char *, which are the MIME types offered by this
     data source.  */
  XLList *mime_types;

  /* List of atoms corresponding to those MIME types, in the same
     order.  */
  XIDList *atom_types;

  /* Number of corresponding MIME types.  */
  int n_mime_types;

  /* The resource associated with this data source.  */
  struct wl_resource *resource;

  /* List of data offers associated with this data source.  */
  DataOffer offers;
};

struct _DataDevice
{
  /* The associated seat.  */
  Seat *seat;

  /* The number of references to this data device.  */
  int refcount;

  /* Linked list of references to this data device.  */
  DataDeviceReference references;
};

/* The data device manager global.  */
static struct wl_global *global_data_device_manager;

/* The current selection.  */
static DataSource *current_selection_data;

/* A sentinel value that means a foreign selection is in use.  */
static DataSource foreign_selection_key;

/* Functions to call to obtain wl_data_offer resources and send
   associated data for foreign selections.  */
static CreateOfferFuncs foreign_selection_functions;

/* Time the foreign selection was changed.  */
static Time foreign_selection_time;

/* When it changed.  */
static uint32_t last_selection_change_serial;

/* Data offer implementation.  */

static void
FreeDataOffer (DataOffer *offer)
{
  /* Mark this offer as invalid by setting the resource's user_data to
     NULL.  */
  if (offer->resource)
    wl_resource_set_user_data (offer->resource, NULL);

  /* Unlink the offer.  */
  offer->last->next = offer->next;
  offer->next->last = offer->last;

  /* Free the offer.  */
  XLFree (offer);
}

static void
Accept (struct wl_client *client, struct wl_resource *resource,
	uint32_t serial, const char *mime_type)
{
  DataOffer *offer;

  offer = wl_resource_get_user_data (resource);

  if (!offer)
    return;

  wl_data_source_send_target (offer->source->resource,
			      mime_type);
}

static void
Receive (struct wl_client *client, struct wl_resource *resource,
	 const char *mime_type, int fd)
{
  DataOffer *offer;

  offer = wl_resource_get_user_data (resource);

  if (!offer)
    {
#ifdef DEBUG
      fprintf (stderr, "wl_client@%p is trying to receive from an outdated"
	       " wl_data_offer@%u\n", client, wl_resource_get_id (resource));
#endif
      close (fd);
      return;
    }

#ifdef DEBUG
  fprintf (stderr, "wl_client@%p is now receiving from wl_data_offer@%u\n",
	   client, wl_resource_get_id (resource));
#endif

  wl_data_source_send_send (offer->source->resource, mime_type, fd);
  close (fd);
}

static void
DestroyDataOffer (struct wl_client *client, struct wl_resource *resource)
{
  wl_resource_destroy (resource);
}

static void
Finish (struct wl_client *client, struct wl_resource *resource)
{
  /* TODO... */
}

static void
DataOfferSetActions (struct wl_client *client, struct wl_resource *resource,
		     uint32_t dnd_actions, uint32_t preferred_action)
{
  /* TODO... */
}

static const struct wl_data_offer_interface wl_data_offer_impl =
  {
    .accept = Accept,
    .receive = Receive,
    .destroy = DestroyDataOffer,
    .finish = Finish,
    .set_actions = DataOfferSetActions,
  };

static void
HandleOfferResourceDestroy (struct wl_resource *resource)
{
  DataOffer *offer;

  offer = wl_resource_get_user_data (resource);

  if (!offer)
    /* The offer was made inert.  */
    return;

  offer->resource = NULL;
  FreeDataOffer (offer);
}

static struct wl_resource *
AddDataOffer (struct wl_client *client, DataSource *source)
{
  DataOffer *offer;
  struct wl_resource *resource;

  resource = wl_resource_create (client, &wl_data_offer_interface,
				 /* 0 means to allocate a new resource
				    ID server-side.  */
				 3, 0);

  if (!resource)
    return NULL;

  offer = XLCalloc (1, sizeof *offer);
  offer->next = source->offers.next;
  offer->last = &source->offers;

  source->offers.next->last = offer;
  source->offers.next = offer;

  offer->resource = resource;
  offer->source = source;

  wl_resource_set_implementation (resource, &wl_data_offer_impl,
				  offer, HandleOfferResourceDestroy);

  return resource;
}


/* Data device and source implementations.  */

/* Forward declaration.  */

static void SendDataOffers (void);

static void
HandleSourceResourceDestroy (struct wl_resource *resource)
{
  DataSource *data_source;
  DataOffer *offer, *last;

  data_source = wl_resource_get_user_data (resource);

  /* If data_source is currently the selection, remove it.  */
  if (data_source == current_selection_data)
    {
      current_selection_data = NULL;

      /* Send the updated data to clients.  */
      SendDataOffers ();
    }

  /* Tell the X selection code that this data source has been
     destroyed.  */
  XLNoteSourceDestroyed (data_source);

  XLListFree (data_source->mime_types, XLFree);
  XIDListFree (data_source->atom_types, NULL);

  /* Make inert and release all data offers on this data source.  */

  offer = data_source->offers.next;

  while (offer != &data_source->offers)
    {
      last = offer;
      offer = offer->next;

      FreeDataOffer (last);
    }

  XLFree (data_source);
}

static const char *
FindAtom (DataSource *source, Atom atom)
{
  XIDList *tem;
  XLList *tem1;

  /* Find the MIME type corresponding to an atom in source.
     Return NULL if the atom is not there.

     source->mime_types must be the same length as
     source->atom_types.  */

  tem = source->atom_types;
  tem1 = source->mime_types;

  for (; tem; tem = tem->next, tem1 = tem1->next)
    {
      if (tem->data == atom)
	return tem1->data;
    }

  return NULL;
}

static void
Offer (struct wl_client *client, struct wl_resource *resource,
       const char *mime_type)
{
  Atom atom;
  DataSource *data_source;
  DataOffer *offer;

  data_source = wl_resource_get_user_data (resource);

  /* It is more efficient to record both atoms and strings in the data
     source, since its contents will be offered to X and Wayland
     clients.  */
  atom = InternAtom (mime_type);

  /* If the type was already offered, simply return.  */
  if (FindAtom (data_source, atom))
    return;

  /* Otherwise, link the atom and the mime type onto the list
     simultaneously.  */
#ifdef DEBUG
  fprintf (stderr, "Offering: %s (X atom: %lu) from wl_data_source@%u\n",
	   mime_type, atom, wl_resource_get_id (resource));
#endif
  data_source->atom_types = XIDListPrepend (data_source->atom_types, atom);
  data_source->mime_types = XLListPrepend (data_source->mime_types,
					   XLStrdup (mime_type));
  data_source->n_mime_types++;

  /* Send the new MIME type to any attached offers.  */

  offer = data_source->offers.next;

  while (offer != &data_source->offers)
    {
      wl_data_offer_send_offer (offer->resource, mime_type);
      offer = offer->next;
    }
}

static void
SetActions (struct wl_client *client, struct wl_resource *resource,
	    uint32_t actions)
{
  /* TODO... */
}

static void
DestroySource (struct wl_client *client, struct wl_resource *resource)
{
  wl_resource_destroy (resource);
}

static const struct wl_data_source_interface wl_data_source_impl =
  {
    .offer = Offer,
    .set_actions = SetActions,
    .destroy = DestroySource,
  };

static void
CreateDataSource (struct wl_client *client, struct wl_resource *resource,
		  uint32_t id)
{
  DataSource *source;

  source = XLSafeMalloc (sizeof *source);

  if (!source)
    {
      wl_resource_post_no_memory (resource);
      return;
    }

  memset (source, 0, sizeof *source);
  source->resource = wl_resource_create (client, &wl_data_source_interface,
					 wl_resource_get_version (resource),
					 id);

  if (!source->resource)
    {
      wl_resource_post_no_memory (resource);
      XLFree (source);
      return;
    }

  source->offers.next = &source->offers;
  source->offers.last = &source->offers;

  wl_resource_set_implementation (source->resource, &wl_data_source_impl,
				  source, HandleSourceResourceDestroy);
}

static void
UpdateSingleReferenceWithForeignOffer (struct wl_client *client,
				       DataDeviceReference *reference)
{
  struct wl_resource *resource;
  Time time;

  time = foreign_selection_time;
  resource = foreign_selection_functions.create_offer (client, time);

  if (!resource)
    return;

  /* Make the data offer known to the client.  */
  wl_data_device_send_data_offer (reference->resource, resource);

  /* Tell the foreign selection provider to send supported resources
     to the client.  */
  foreign_selection_functions.send_offers (resource, time);

  /* Finally, tell the client that the offer is a selection.  */
  wl_data_device_send_selection (reference->resource, resource);
}

static void
UpdateForSingleReference (DataDeviceReference *device)
{
  struct wl_resource *resource;
  struct wl_client *client;
  XLList *type;

  if (!current_selection_data)
    {
      wl_data_device_send_selection (device->resource,
				     NULL);
      return;
    }

  client = wl_resource_get_client (device->resource);

  if (current_selection_data == &foreign_selection_key)
    {
      /* This means a foreign selection is in use.  */
      UpdateSingleReferenceWithForeignOffer (client, device);

      return;
    }

  resource = AddDataOffer (client, current_selection_data);

  if (!resource)
    return;

  /* First, introduce the data offer to the client.  */
  wl_data_device_send_data_offer (device->resource, resource);

  /* Send all the offered MIME types.  */

  type = current_selection_data->mime_types;

  for (; type; type = type->next)
    wl_data_offer_send_offer (resource, type->data);

  /* Finally, tell the client it is a selection.  */
  wl_data_device_send_selection (device->resource, resource);
}

static void
SendDataOffersForDevice (DataDevice *device)
{
  DataDeviceReference *reference;
  struct wl_client *client;

  reference = device->references.next;

  while (reference != &device->references)
    {
      client = wl_resource_get_client (reference->resource);

      if (XLSeatIsClientFocused (device->seat, client))
	UpdateForSingleReference (reference);

      reference = reference->next;
    }
}

static void
SendDataOffers (void)
{
  XLList *seat;
  DataDevice *device;

  for (seat = live_seats; seat; seat = seat->next)
    {
      device = XLSeatGetDataDevice (seat->data);

      if (device)
	SendDataOffersForDevice (device);
    }
}

static void
DestroyReference (DataDeviceReference *reference)
{
  reference->next->last = reference->last;
  reference->last->next = reference->next;

  XLFree (reference);
}

static void
ReleaseReferences (DataDevice *device)
{
  DataDeviceReference *reference;

  reference = device->references.next;

  while (reference != &device->references)
    {
      reference->device = NULL;
      reference = reference->next;
    }
}

static void
DestroyBacking (DataDevice *device)
{
  if (--device->refcount)
    return;

  ReleaseReferences (device);
  XLFree (device);
}

static DataDevice *
GetDataDeviceInternal (Seat *seat)
{
  DataDevice *device;

  device = XLSeatGetDataDevice (seat);

  if (!device)
    {
      device = XLCalloc (1, sizeof *device);
      device->seat = seat;
      device->references.next = &device->references;
      device->references.last = &device->references;

      XLSeatSetDataDevice (seat, device);
    }

  return device;
}

static DataDeviceReference *
AddReferenceTo (DataDevice *device, struct wl_resource *resource)
{
  DataDeviceReference *reference;

  reference = XLCalloc (1, sizeof *reference);
  reference->next = device->references.next;
  reference->last = &device->references;
  reference->resource = resource;

  device->references.next->last = reference;
  device->references.next = reference;

  reference->device = device;

  return reference;
}

static void
StartDrag (struct wl_client *client, struct wl_resource *resource,
	   struct wl_resource *source_resource,
	   struct wl_resource *origin_resource,
	   struct wl_resource *icon_resource, uint32_t serial)
{
  /* TODO... */
}

static void
SetSelection (struct wl_client *client, struct wl_resource *resource,
	      struct wl_resource *source_resource, uint32_t serial)
{
  DataSource *source;
  DataDeviceReference *device;

#ifdef DEBUG
  if (source_resource)
    fprintf (stderr, "wl_client@%p is setting the selection to "
	     "wl_data_source@%u, at %u\n",
	     client, wl_resource_get_id (source_resource), serial);
#endif

  if (serial < last_selection_change_serial)
    {
#ifdef DEBUG
      fprintf (stderr, "wl_client@%p could not set the selection, "
	       "because the selection changed.  (%u < %u)\n",
	       client, serial, last_selection_change_serial);
#endif
      return;
    }

  device = wl_resource_get_user_data (resource);

  if (!device->device || !device->device->seat)
    /* This device is inert, since the seat has been deleted.  */
    return;

  /* Set the last selection change serial.  This enables us to avoid
     race conditions caused by multiple clients simultaneously setting
     the clipboard according to different events, by keeping track of
     which event is newer.  */
  last_selection_change_serial = serial;

  if (source_resource)
    source = wl_resource_get_user_data (source_resource);
  else
    source = NULL;

  /* Try to own the X selection.  If it fails, refrain from changing
     the current selection data.  */
  if (!XLNoteLocalSelection (device->device->seat, source))
    return;

  if (current_selection_data && current_selection_data != source)
    {
      /* If the current selection data is already set, cancel it.  */
      if (current_selection_data != &foreign_selection_key)
	wl_data_source_send_cancelled (current_selection_data->resource);
    }

  if (current_selection_data != source)
    {
      current_selection_data = source;

      /* Create data offer objects for the new selection data and send
	 it to clients.  */
      SendDataOffers ();
    }
}

static void
Release (struct wl_client *client, struct wl_resource *resource)
{
  wl_resource_destroy (resource);
}

static const struct wl_data_device_interface wl_data_device_impl =
  {
    .start_drag = StartDrag,
    .set_selection = SetSelection,
    .release = Release,
  };

static void
HandleDeviceResourceDestroy (struct wl_resource *resource)
{
  DataDeviceReference *reference;

  reference = wl_resource_get_user_data (resource);
  DestroyReference (reference);
}

static void
GetDataDevice (struct wl_client *client, struct wl_resource *resource,
	       uint32_t id, struct wl_resource *seat_resource)
{
  struct wl_resource *device_resource;
  Seat *seat;
  DataDevice *device;
  DataDeviceReference *reference;

  device_resource = wl_resource_create (client, &wl_data_device_interface,
					wl_resource_get_version (resource),
					id);

  if (!device_resource)
    {
      wl_resource_post_no_memory (resource);
      return;
    }

  seat = wl_resource_get_user_data (seat_resource);
  device = GetDataDeviceInternal (seat);
  reference = AddReferenceTo (device, device_resource);

  wl_resource_set_implementation (device_resource, &wl_data_device_impl,
				  reference, HandleDeviceResourceDestroy);
}

static struct wl_data_device_manager_interface wl_data_device_manager_impl =
  {
    .create_data_source = CreateDataSource,
    .get_data_device = GetDataDevice,
  };

static void
HandleBind (struct wl_client *client, void *data,
	    uint32_t version, uint32_t id)
{
  struct wl_resource *resource;

  resource
    = wl_resource_create (client, &wl_data_device_manager_interface,
			  version, id);

  if (!resource)
    {
      wl_client_post_no_memory (client);
      return;
    }

  wl_resource_set_implementation (resource,
				  &wl_data_device_manager_impl,
				  NULL, NULL);
}

void
XLInitDataDevice (void)
{
  global_data_device_manager
    = wl_global_create (compositor.wl_display,
			&wl_data_device_manager_interface,
			3, NULL, HandleBind);
}

void
XLRetainDataDevice (DataDevice *device)
{
  device->refcount++;
}

void
XLReleaseDataDevice (DataDevice *device)
{
  DestroyBacking (device);
}

void
XLDataDeviceClearSeat (DataDevice *device)
{
  device->seat = NULL;
}

void
XLDataDeviceHandleFocusChange (DataDevice *device)
{
  SendDataOffersForDevice (device);
}

void
XLSetForeignSelection (Time time, CreateOfferFuncs functions)
{
  uint32_t serial;

  if (time < foreign_selection_time)
    return;

  serial = wl_display_next_serial (compositor.wl_display);

  /* Use this serial to prevent clients from changing the selection
     again until the next event is sent.  */
  last_selection_change_serial = serial;

  foreign_selection_time = time;
  foreign_selection_functions = functions;

  if (current_selection_data
      && current_selection_data != &foreign_selection_key)
    wl_data_source_send_cancelled (current_selection_data->resource);

  /* Use a special value of current_selection_data to mean that
     foreign selections are in use.  */
  current_selection_data = &foreign_selection_key;

  /* Send new data offers to current clients.  */
  SendDataOffers ();
}

void
XLClearForeignSelection (Time time)
{
  if (time < foreign_selection_time)
    return;

  if (current_selection_data == &foreign_selection_key)
    {
      current_selection_data = NULL;

      SendDataOffers ();
    }

  foreign_selection_time = time;
}

int
XLDataSourceTargetCount (DataSource *source)
{
  return source->n_mime_types;
}

void
XLDataSourceGetTargets (DataSource *source, Atom *targets)
{
  int i;
  XIDList *list;

  list = source->atom_types;

  for (i = 0; i < source->n_mime_types; ++i)
    {
      targets[i] = list->data;
      list = list->next;
    }
}

struct wl_resource *
XLResourceFromDataSource (DataSource *source)
{
  return source->resource;
}

Bool
XLDataSourceHasAtomTarget (DataSource *source, Atom target)
{
  XIDList *list;

  for (list = source->atom_types; list; list = list->next)
    {
      if (list->data == target)
	return True;
    }

  return False;
}

Bool
XLDataSourceHasTarget (DataSource *source, const char *mime_type)
{
  XLList *list;

  for (list = source->mime_types; list; list = list->next)
    {
      if (!strcmp (list->data, mime_type))
	return True;
    }

  return False;
}

void
XLDataDeviceMakeOffers (Seat *seat, DndOfferFuncs funcs, Surface *surface,
			int x, int y)
{
  DataDevice *device;
  DataDeviceReference *reference;
  struct wl_resource *resource;
  struct wl_client *client;
  int version;
  uint32_t serial;

  device = XLSeatGetDataDevice (seat);
  client = wl_resource_get_client (surface->resource);
  reference = device->references.next;
  serial = wl_display_next_serial (compositor.wl_display);

  while (reference != &device->references)
    {
      /* If this reference to the data device belongs to the client,
	 continue.  */
      if (wl_resource_get_client (reference->resource)
	  == wl_resource_get_client (surface->resource))
	{
	  version = wl_resource_get_version (reference->resource);

	  /* Create the offer.  */
	  resource = funcs.create (client, version);

	  if (resource)
	    {
	      /* Actually send the data offer to the client.  */
	      wl_data_device_send_data_offer (reference->resource,
					      resource);

	      /* And data offers.  */
	      funcs.send_offers (resource);

	      /* And send the entry event.  */
	      wl_data_device_send_enter (reference->resource,
					 serial, surface->resource,
					 wl_fixed_from_double (x),
					 wl_fixed_from_double (y),
					 resource);
	    }
	}

      reference = reference->next;
    }
}

void
XLDataDeviceSendMotion (Seat *seat, Surface *surface,
			int x, int y, Time time)
{
  DataDevice *device;
  DataDeviceReference *reference;

  device = XLSeatGetDataDevice (seat);

  if (!device)
    /* No data device has been created for this seat yet.  */
    return;

  reference = device->references.next;

  while (reference != &device->references)
    {
      /* If this reference to the data device belongs to the client,
	 continue.  */
      if (wl_resource_get_client (reference->resource)
	  == wl_resource_get_client (surface->resource))
	wl_data_device_send_motion (reference->resource, time,
				    wl_fixed_from_double (x),
				    wl_fixed_from_double (y));

      reference = reference->next;
    }
}

void
XLDataDeviceSendLeave (Seat *seat, Surface *surface)
{
  DataDevice *device;
  DataDeviceReference *reference;

  device = XLSeatGetDataDevice (seat);

  if (!device)
    /* No data device has been created for this seat yet.  */
    return;

  reference = device->references.next;

  while (reference != &device->references)
    {
      /* If this reference to the data device belongs to the client,
	 continue.  */
      if (wl_resource_get_client (reference->resource)
	  == wl_resource_get_client (surface->resource))
	wl_data_device_send_leave (reference->resource);

      reference = reference->next;
    }
}

void
XLDataDeviceSendDrop (Seat *seat, Surface *surface)
{
  DataDevice *device;
  DataDeviceReference *reference;

  device = XLSeatGetDataDevice (seat);

  if (!device)
    /* No data device has been created for this seat yet.  */
    return;

  reference = device->references.next;

  while (reference != &device->references)
    {
      /* If this reference to the data device belongs to the client,
	 continue.  */
      if (wl_resource_get_client (reference->resource)
	  == wl_resource_get_client (surface->resource))
	wl_data_device_send_drop (reference->resource);

      reference = reference->next;
    }
}
