/* Wayland compositor running on top of an X serer.

Copyright (C) 2022 to various contributors.

This file is part of 12to11.

12to11 is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

12to11 is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with 12to11.  If not, see <https://www.gnu.org/licenses/>.  */

#include <stdio.h>
#include <string.h>

#include "compositor.h"

enum
  {
    XdndProtocolVersion = 5,
  };

typedef struct _DndState DndState;

struct _DndState
{
  /* The source window.  */
  Window source_window;

  /* The target window.  */
  Window target_window;

  /* The protocol version in use.  */
  int proto;

  /* The seat that is being used.  */
  Seat *seat;

  /* The key for the seat destruction callback.  */
  void *seat_callback;

  /* Array of selection targets, which are MIME types in the Xdnd
     protocol, making our interaction with Wayland clients very
     convenient.  */
  char **targets;

  /* Number of targets in that array.  */
  int ntargets;

  /* Monotonically increasing counter.  */
  unsigned int serial;

  /* Whether or not non-default values should be used to respond to
     drag-and-drop events.  */
  Bool respond;

  /* The struct wl_resource (s) associated with this drag and drop
     operation.  */
  XLList *resources;

  /* The surface associated with this drag and drop session.  */
  Surface *surface;

  /* The destroy callback associated with that surface.  */
  DestroyCallback *callback;

  /* The source action mask.  */
  uint32_t source_actions;

  /* The supported action and preferred action.  */
  uint32_t supported_actions, preferred_action;

  /* The chosen DND action.  */
  uint32_t used_action;

  /* Whether or not something was accepted.  */
  Bool accepted;

  /* Whether or not the transfer finished.  */
  Bool finished;

  /* Whether or not the drop has already happened.  */
  Bool dropped;

  /* The timestamp to use for accessing selection data.  */
  Time timestamp;

  /* The toplevel or child surface the pointer is currently
     inside.  */
  Surface *child;

  /* The unmap callback for that child.  */
  UnmapCallback *unmap_callback;

  /* The version of the XDND protocol being used.  */
  int version;
};

/* The global drag-and-drop state.  */
DndState dnd_state;

/* Forward declaration.  */

static void FinishDndEntry (void);

static Seat *
AssignSeat (void)
{
  /* Since the XDND protocol doesn't provide any way to determine the
     seat a drag-and-drop operation is originating from, simply return
     the first seat to be created.  */

  if (live_seats)
    return live_seats->data;

  return NULL;
}

static void
HandleSeatDestroy (void *data)
{
  dnd_state.seat = NULL;
  dnd_state.seat_callback = NULL;

  /* Since the seat has been destroyed, finish the drag and drop
     operation.  */
  FinishDndEntry ();
}

static uint32_t
TranslateAction (Atom action)
{
  if (action == XdndActionCopy)
    return WL_DATA_DEVICE_MANAGER_DND_ACTION_COPY;

  if (action == XdndActionMove)
    return WL_DATA_DEVICE_MANAGER_DND_ACTION_MOVE;

  if (action == XdndActionAsk)
    return WL_DATA_DEVICE_MANAGER_DND_ACTION_ASK;

  /* Wayland doesn't have an equivalent to XdndActionPrivate, so fall
     back to copy.  */
  if (action == XdndActionPrivate)
    return WL_DATA_DEVICE_MANAGER_DND_ACTION_COPY;

  /* Otherwise, return None.  */
  return WL_DATA_DEVICE_MANAGER_DND_ACTION_NONE;
}

/* Forward declarations.  */

static void SendStatus (void);
static void RespondToDndDrop (void);

static void
Accept (struct wl_client *client, struct wl_resource *resource,
	uint32_t serial, const char *mime_type)
{
  uint32_t sc;

  sc = (intptr_t) wl_resource_get_user_data (resource);

  if (sc < dnd_state.serial
      || dnd_state.source_window == None)
    /* This data offer is out of date.  */
    return;

  if (wl_resource_get_version (resource) <= 2)
    /* In version 2 and below, this doesn't affect anything.  */
    return;

  if (!mime_type)
    {
      if (dnd_state.accepted)
	/* The accepted state changed.  */
	SendStatus ();

      dnd_state.accepted = False;
    }
  else
    {
      if (!dnd_state.accepted)
	/* The accepted state changed.  */
	SendStatus ();

      dnd_state.accepted = True;
    }
}

static void
Receive (struct wl_client *client, struct wl_resource *resource,
	 const char *mime_type, int fd)
{
  uint32_t serial;

  serial = (intptr_t) wl_resource_get_user_data (resource);

  if (serial < dnd_state.serial
      || dnd_state.source_window == None)
    {
      /* This data offer is out of date.  */
      close (fd);
      return;
    }

  XLReceiveDataFromSelection (dnd_state.timestamp, XdndSelection,
			      InternAtom (mime_type), fd);
}

static void
Destroy (struct wl_client *client, struct wl_resource *resource)
{
  wl_resource_destroy (resource);
}

static void
Finish (struct wl_client *client, struct wl_resource *resource)
{
  uint32_t serial;

  serial = (intptr_t) wl_resource_get_user_data (resource);

  if (serial < dnd_state.serial
      || !dnd_state.used_action
      || !dnd_state.accepted
      || dnd_state.finished)
    {
      wl_resource_post_error (resource, WL_DATA_OFFER_ERROR_INVALID_FINISH,
			      "finish called at inopportune moment");
      return;
    }

  dnd_state.finished = True;

  /* If XdndDrop was received, send the XdndFinished message.  */
  if (dnd_state.dropped)
    RespondToDndDrop ();
}

static Atom
ConvertAction (uint32_t action)
{
  if (action == WL_DATA_DEVICE_MANAGER_DND_ACTION_COPY)
    return XdndActionCopy;

  if (action == WL_DATA_DEVICE_MANAGER_DND_ACTION_MOVE)
    return XdndActionMove;

  if (action == WL_DATA_DEVICE_MANAGER_DND_ACTION_ASK)
    return XdndActionAsk;

  return None;
}

static void
SendStatus (void)
{
  XEvent event;

  if (dnd_state.dropped)
    return;

  memset (&event, 0, sizeof event);
  event.xclient.type = ClientMessage;
  event.xclient.window = dnd_state.source_window;
  event.xclient.message_type = XdndStatus;
  event.xclient.format = 32;

  event.xclient.data.l[0] = dnd_state.target_window;

  if (dnd_state.respond)
    {
      if ((dnd_state.used_action
	   != WL_DATA_DEVICE_MANAGER_DND_ACTION_NONE)
	  && dnd_state.accepted)
	event.xclient.data.l[1] = 1;

      if (dnd_state.version >= 3)
	event.xclient.data.l[4] = ConvertAction (dnd_state.used_action);
      else
	/* The version of the data device manager protocol spoken by
	   the client doesn't support actions.  Use XdndActionPrivate.  */
	event.xclient.data.l[4] = XdndActionPrivate;
    }

  CatchXErrors ();
  XSendEvent (compositor.display, dnd_state.source_window,
	      False, NoEventMask, &event);
  UncatchXErrors (NULL);
}

static void
UpdateUsedAction (void)
{
  uint32_t intersection, old;
  XLList *list;

  old = dnd_state.used_action;

  /* First, see if the preferred action is supported.  If it is,
     simply use it.  */
  if (dnd_state.source_actions & dnd_state.preferred_action)
    dnd_state.used_action = dnd_state.preferred_action;
  else
    {
      intersection = (dnd_state.supported_actions
		      & dnd_state.source_actions);

      /* Now, try the following actions in order.  */
      if (intersection & WL_DATA_DEVICE_MANAGER_DND_ACTION_COPY)
	dnd_state.used_action = WL_DATA_DEVICE_MANAGER_DND_ACTION_COPY;
      else if (intersection & WL_DATA_DEVICE_MANAGER_DND_ACTION_MOVE)
	dnd_state.used_action = WL_DATA_DEVICE_MANAGER_DND_ACTION_MOVE;
      else if (intersection & WL_DATA_DEVICE_MANAGER_DND_ACTION_ASK)
	dnd_state.used_action = WL_DATA_DEVICE_MANAGER_DND_ACTION_ASK;
      else
	dnd_state.used_action = WL_DATA_DEVICE_MANAGER_DND_ACTION_NONE;
    }

  /* Send the updated action to clients if it changed.  */
  if (old != dnd_state.used_action)
    {
      for (list = dnd_state.resources; list; list = list->next)
	{
	  if (wl_resource_get_version (list->data) >= 3)
	    wl_data_offer_send_action (list->data,
				       dnd_state.used_action);
	}
    }

  /* Send an XdndStatus if the action changed.  */
  SendStatus ();
}

static void
SetActions (struct wl_client *client, struct wl_resource *resource,
	    uint32_t dnd_actions, uint32_t preferred_action)
{
  uint32_t serial;

  serial = (intptr_t) wl_resource_get_user_data (resource);

  if (serial < dnd_state.serial
      || !dnd_state.source_window)
    /* This data offer is out of date.  */
    return;

  if (dnd_actions & ~(WL_DATA_DEVICE_MANAGER_DND_ACTION_COPY
		      | WL_DATA_DEVICE_MANAGER_DND_ACTION_MOVE
		      | WL_DATA_DEVICE_MANAGER_DND_ACTION_ASK)
      || (preferred_action != WL_DATA_DEVICE_MANAGER_DND_ACTION_ASK
	  && preferred_action != WL_DATA_DEVICE_MANAGER_DND_ACTION_MOVE
	  && preferred_action != WL_DATA_DEVICE_MANAGER_DND_ACTION_COPY
	  && preferred_action != WL_DATA_DEVICE_MANAGER_DND_ACTION_NONE))
    wl_resource_post_error (resource, WL_DATA_OFFER_ERROR_INVALID_ACTION,
			    "invalid action or action mask among: %u %u",
			    dnd_actions, preferred_action);

  /* Otherwise, update the DND state with the supported action.  */
  dnd_state.supported_actions = dnd_actions;
  dnd_state.preferred_action = preferred_action;

  /* And send the updated state.  */
  UpdateUsedAction ();
}

static const struct wl_data_offer_interface wl_data_offer_impl =
  {
    .accept = Accept,
    .receive = Receive,
    .destroy = Destroy,
    .finish = Finish,
    .set_actions = SetActions,
  };

static void
HandleResourceDestroy (struct wl_resource *resource)
{
  uint32_t serial;

  serial = (intptr_t) wl_resource_get_user_data (resource);

  if (serial >= dnd_state.serial
      && dnd_state.source_window != None)
    {
      /* Send XdndFinish if it hasn't already been sent.  Since the
	 resource has been destroyed without previously completing,
	 signal an error if its version is 3 or later.  */

      if (wl_resource_get_version (resource) >= 3)
	dnd_state.accepted = False;

      if (dnd_state.dropped)
	RespondToDndDrop ();

      /* Remove the resource from the resource list.  */
      dnd_state.resources = XLListRemove (dnd_state.resources,
					  resource);

      /* If there are no more resources, finish the drag and drop
	 operation.  Note that this might've already been done by
	 RespondToDndDrop, but it is safe to call FinishDndEntry
	 twice.  */
      if (!dnd_state.resources)
	FinishDndEntry ();
    }
}

static struct wl_resource *
CreateOffer (struct wl_client *client, int version)
{
  struct wl_resource *resource;

  resource = wl_resource_create (client, &wl_data_offer_interface,
				 version, 0);

  if (!resource)
    return NULL;

  wl_resource_set_implementation (resource, &wl_data_offer_impl,
				  (void *) (intptr_t) dnd_state.serial,
				  HandleResourceDestroy);
  dnd_state.resources = XLListPrepend (dnd_state.resources,
				       resource);

  /* If version <= 2, then the drag-and-drop operation should always
     be accepted, no matter whether or not accept is called.  */
  if (version <= 2)
    dnd_state.accepted = True;

  if (!dnd_state.version || dnd_state.version > version)
    dnd_state.version = version;

  return resource;
}

static void
SendOffers (struct wl_resource *resource)
{
  int i;

  for (i = 0; i < dnd_state.ntargets; ++i)
    {
      if (dnd_state.targets[i])
	wl_data_offer_send_offer (resource,
				  dnd_state.targets[i]);
    }
}

static void
FinishDndEntry (void)
{
  int i;

  if (dnd_state.seat && dnd_state.resources
      /* Don't send leave if a drop already happened.  */
      && !dnd_state.dropped)
    XLDataDeviceSendLeave (dnd_state.seat, dnd_state.surface);

  dnd_state.source_window = None;
  dnd_state.target_window = None;
  dnd_state.surface = NULL;
  dnd_state.proto = 0;

  if (dnd_state.callback)
    XLSurfaceCancelRunOnFree (dnd_state.callback);
  dnd_state.callback = NULL;

  if (dnd_state.seat)
    XLSeatCancelDestroyListener (dnd_state.seat_callback);
  dnd_state.seat = NULL;
  dnd_state.seat_callback = NULL;

  if (dnd_state.child)
    XLSurfaceCancelUnmapCallback (dnd_state.unmap_callback);
  dnd_state.child = NULL;
  dnd_state.unmap_callback = NULL;

  for (i = 0; i < dnd_state.ntargets; ++i)
    {
      if (dnd_state.targets[i])
	XFree (dnd_state.targets[i]);
    }

  XLFree (dnd_state.targets);
  dnd_state.ntargets = 0;
  dnd_state.targets = NULL;
  dnd_state.source_actions = 0;
  dnd_state.supported_actions = 0;
  dnd_state.preferred_action = 0;
  dnd_state.used_action = 0;
  dnd_state.version = 0;
  dnd_state.accepted = False;
  dnd_state.finished = False;
  dnd_state.dropped = False;
  dnd_state.timestamp = CurrentTime;

  /* The resources are not destroyed, since the client will do that
     later.  */
  XLListFree (dnd_state.resources, NULL);
  dnd_state.resources = NULL;
}

static void
RespondToDndDrop (void)
{
  XEvent event;

  memset (&event, 0, sizeof event);
  event.xclient.type = ClientMessage;
  event.xclient.window = dnd_state.source_window;
  event.xclient.message_type = XdndFinished;
  event.xclient.format = 32;

  event.xclient.data.l[0] = dnd_state.target_window;

  if (dnd_state.proto >= 5
      && dnd_state.used_action && dnd_state.accepted
      && dnd_state.seat && dnd_state.respond)
    {
      /* This determines whether or not the drag and drop operation
	 was accepted.  */
      event.xclient.data.l[1] = 1;

      if (dnd_state.version >= 3)
	/* And this specifies the action that was really taken.  */
	event.xclient.data.l[2] = ConvertAction (dnd_state.used_action);
      else
	/* The version of the data device manager protocol spoken by
	   the client doesn't support actions.  Use XdndActionPrivate.  */
	event.xclient.data.l[2] = XdndActionPrivate;
    }

  CatchXErrors ();
  XSendEvent (compositor.display, dnd_state.source_window,
	      False, NoEventMask, &event);
  UncatchXErrors (NULL);

  /* Now that XdndFinished has been sent, the drag and drop operation
     is complete.  */
  FinishDndEntry ();
}

static void
HandleSurfaceDestroy (void *data)
{
  dnd_state.surface = NULL;
  dnd_state.callback = NULL;
}

static void
HandleDndEntry (Surface *target, Window source, Atom *targets,
		int ntargets, int proto)
{
  int i;
  char **names;

  if (dnd_state.source_window)
    {
      fprintf (stderr, "XdndEnter received while a drag-and-drop operation"
	       " is in progress; overriding current drag-and-drop operation\n");
      FinishDndEntry ();
    }

  dnd_state.proto = proto;
  dnd_state.source_window = source;
  dnd_state.surface = target;
  dnd_state.callback = XLSurfaceRunOnFree (dnd_state.surface,
					   HandleSurfaceDestroy, NULL);

  /* Retrieve the atoms inside the targets list.  */
  names = XLCalloc (ntargets, sizeof *names);

  CatchXErrors ();
  XGetAtomNames (compositor.display, targets,
		 ntargets, names);
  UncatchXErrors (NULL);

  /* Enter the names of the targets into the atom table so that they
     can be interned without roundtrips in the future.  */
  for (i = 0; i < ntargets; ++i)
    {
      if (names[i])
	ProvideAtom (names[i], targets[i]);
    }

  /* Find a seat to use for this drag-and-drop operation.  */
  dnd_state.seat = AssignSeat ();

  /* If a seat was found, listen for its destruction.  After the
     initiating seat is destroyed (or if none was found), we reply to
     all future drag-and-drop messages with dummy values.  */
  if (dnd_state.seat)
    dnd_state.seat_callback = XLSeatRunOnDestroy (dnd_state.seat,
						  HandleSeatDestroy,
						  NULL);

  /* Initialize available data types from the atom names.  */
  dnd_state.targets = names;
  dnd_state.ntargets = ntargets;

  /* Initialize other drag-and-drop state.  */
  dnd_state.respond = False;

  /* There shouldn't be any leftovers from the last session.  */
  XLAssert (dnd_state.resources == NULL);

  /* Initialize the target window.  */
  dnd_state.target_window = XLWindowFromSurface (target);

  /* Increase the state counter to make all out-of-date data offers
     invalid.  */
  dnd_state.serial++;
}

static Atom *
ReadXdndTypeList (Window window, int *nitems_return)
{
  Atom actual_type;
  int rc, actual_format;
  unsigned long nitems, bytes_remaining;
  unsigned char *tmp_data;

  tmp_data = NULL;

  CatchXErrors ();
  rc = XGetWindowProperty (compositor.display, window,
			   XdndTypeList, 0, LONG_MAX,
			   False, XA_ATOM, &actual_type,
			   &actual_format, &nitems,
			   &bytes_remaining, &tmp_data);
  if (UncatchXErrors (NULL) || rc != Success || actual_format != 32
      || !tmp_data || actual_type != XA_ATOM || nitems < 1)
    {
      if (tmp_data)
	XFree (tmp_data);

      return NULL;
    }

  *nitems_return = nitems;
  return (Atom *) tmp_data;
}

static Bool
HandleXdndEnterEvent (Surface *surface, XEvent *event)
{
  Atom *targets;
  Atom builtin[3];
  int ntargets, proto;

  if (event->xclient.data.l[1] & 1)
    /* There are more than 3 targets; retrieve them from the
       XdndTypeList property.  */
    targets = ReadXdndTypeList (event->xclient.data.l[0],
				&ntargets);
  else
    {
      /* Otherwise, the first three properties contain the selection
	 targets.  */
      targets = builtin;
      ntargets = 0;

      if (event->xclient.data.l[2])
	builtin[ntargets++] = event->xclient.data.l[2];

      if (event->xclient.data.l[3])
	builtin[ntargets++] = event->xclient.data.l[3];

      if (event->xclient.data.l[4])
	builtin[ntargets++] = event->xclient.data.l[4];
    }

  if (!targets)
    /* For some reason we failed to retrieve XdndTypeList.  Ignore the
       XdndEnter event.  */
    return True;

  proto = MIN (event->xclient.data.l[1] >> 24,
	       XdndProtocolVersion);

  HandleDndEntry (surface, event->xclient.data.l[0],
		  targets, ntargets, proto);

  if (event->xclient.data.l[1] & 1)
    /* Now, free the type list, which was allocated by Xlib.  */
    XFree (targets);

  return True;
}

static void
HandleChildUnmap (void *data)
{
  /* The child was unmapped.  */

  if (dnd_state.seat)
    XLDataDeviceSendLeave (dnd_state.seat, dnd_state.child);
  XLSurfaceCancelUnmapCallback (dnd_state.unmap_callback);

  dnd_state.child = NULL;
  dnd_state.unmap_callback = NULL;

  /* Free our record of the data offers introduced at entry time; it
     is assumed that the client will delete them too.  */
  XLListFree (dnd_state.resources, NULL);
  dnd_state.resources = NULL;
}

static Bool
HandleMotion (Surface *toplevel, int x, int y, uint32_t action,
	      int *x_out, int *y_out)
{
  Subcompositor *subcompositor;
  View *view;
  int x_off, y_off;
  Surface *child;
  DndOfferFuncs funcs;
  XLList *tem;

  subcompositor = ViewGetSubcompositor (toplevel->view);

  /* Find the view underneath the subcompositor.  */
  view = SubcompositorLookupView (subcompositor, x, y,
				  &x_off, &y_off);
  child = ViewGetData (view);

  /* Compute the surface-relative coordinates and scale them.  */
  *x_out = (x - x_off) / global_scale_factor;
  *y_out = (y - y_off) / global_scale_factor;

  if (dnd_state.child == child)
    /* If nothing changed, don't do anything.  */
    return False;

  /* If the pointer was previously in a different surface, leave
     it.  */
  if (dnd_state.child)
    {
      XLDataDeviceSendLeave (dnd_state.seat,
			     dnd_state.child);
      XLSurfaceCancelUnmapCallback (dnd_state.unmap_callback);

      dnd_state.child = NULL;
      dnd_state.unmap_callback = NULL;

      /* Free our record of the data offers introduced at entry time;
	 it is assumed that the client will delete them too.  */
      XLListFree (dnd_state.resources, NULL);
      dnd_state.resources = NULL;
      dnd_state.used_action = 0;
      dnd_state.preferred_action = 0;
      dnd_state.supported_actions = 0;
      dnd_state.accepted = False;

      if (dnd_state.version <= 2)
	dnd_state.accepted = True;
    }

  /* Now, enter the new surface.  */
  if (child)
    {
      dnd_state.child = child;
      dnd_state.unmap_callback
	= XLSurfaceRunAtUnmap (child, HandleChildUnmap, NULL);
      funcs.create = CreateOffer;
      funcs.send_offers = SendOffers;

      /* Create the offers and send data to the clients.  */
      XLDataDeviceMakeOffers (dnd_state.seat, funcs, child, *x_out,
			      *y_out);
      /* Send source actions to each resource created.  */
      for (tem = dnd_state.resources; tem; tem = tem->next)
	{
	  if (wl_resource_get_version (tem->data) >= 3)
	    wl_data_offer_send_source_actions (tem->data, action);
	}

      /* Now compute whether or not we should respond with actual
	 values.  */
      dnd_state.respond = (dnd_state.resources != NULL);
    }

  return child != NULL;
}

static uint32_t
ReadDndActionList (Window window)
{
  uint32_t mask;
  Atom actual_type, *atoms;
  int rc, actual_format;
  unsigned long nitems, bytes_remaining, i;
  unsigned char *tmp_data;

  mask = WL_DATA_DEVICE_MANAGER_DND_ACTION_NONE;
  tmp_data = NULL;

  CatchXErrors ();
  rc = XGetWindowProperty (compositor.display, window,
			   XdndActionList, 0, LONG_MAX,
			   False, XA_ATOM, &actual_type,
			   &actual_format, &nitems,
			   &bytes_remaining, &tmp_data);
  if (UncatchXErrors (NULL) || rc != Success || actual_format != 32
      || !tmp_data || actual_type != XA_ATOM || nitems < 1)
    {
      if (tmp_data)
	XFree (tmp_data);

      return WL_DATA_DEVICE_MANAGER_DND_ACTION_NONE;
    }

  atoms = (Atom *) tmp_data;

  for (i = 0; i < nitems; ++i)
    mask |= TranslateAction (atoms[i]);

  return mask;
}

static Bool
HandleXdndPositionEvent (Surface *surface, XEvent *event)
{
  int root_x, root_y, x, y;
  Window child;
  XLList *tem;
  uint32_t action;
  Bool sent_actions;

  /* TODO: handle subsurfaces.  */

  if (event->xclient.data.l[0] != dnd_state.source_window)
    /* The message is coming from the wrong window, or drag and drop
       has not yet been set up.  */
    return True;

  if (surface != dnd_state.surface)
    /* This message is being delivered to the wrong surface.  */
    return True;

  /* Extract the root X and root Y from the event.  */
  root_x = event->xclient.data.l[2] >> 16;
  root_y = event->xclient.data.l[2] & 0xffff;

  /* Translate the coordinates to the surface's window.  */
  XTranslateCoordinates (compositor.display,
			 DefaultRootWindow (compositor.display),
			 XLWindowFromSurface (surface),
			 root_x, root_y, &x, &y, &child);

  /* Compute the action.  TODO: handle multiple actions.  */
  action = TranslateAction (event->xclient.data.l[4]);

  /* Handle mouse motion.  */
  sent_actions = HandleMotion (surface, x, y, action, &x, &y);

  if (action == WL_DATA_DEVICE_MANAGER_DND_ACTION_ASK)
    {
      if (!(action & dnd_state.source_actions))
	/* Fetch the list of available actions, and give that to the
	   client along with the regular action list, if XdndActionAsk
	   is being specified for the first time.  */
	action |= ReadDndActionList (dnd_state.source_window);
      else
	/* Otherwise, preserve the action list that was already
	   read.  */
	action |= dnd_state.source_actions;
    }
  /* Send actions from all data offers.  */
  if (dnd_state.resources && !sent_actions)
    {
      /* If action is different from the current source action, send
	 the new source action to the client.  */

      if (action != dnd_state.source_actions)
	{
	  /* Send source actions to each resource created.  */
	  for (tem = dnd_state.resources; tem; tem = tem->next)
	    {
	      if (wl_resource_get_version (tem->data) >= 3)
		wl_data_offer_send_source_actions (tem->data, action);
	    }

	  /* Update the chosen action.  */
	  UpdateUsedAction ();
	}
    }

  dnd_state.source_actions = action;
  dnd_state.timestamp = event->xclient.data.l[3];

  if (dnd_state.seat && dnd_state.child)
    XLDataDeviceSendMotion (dnd_state.seat, surface,
			    /* l[3] is the timestamp of the
			       movement.  */
			    x, y, event->xclient.data.l[3]);

  /* Send an XdndStatus event in response.  */
  SendStatus ();

  return True;
}

static Bool
HandleXdndLeaveEvent (Surface *surface, XEvent *event)
{
  if (event->xclient.data.l[0] != dnd_state.source_window)
    /* The message is coming from the wrong window, or drag and drop
       has not yet been set up.  */
    return True;

  if (surface != dnd_state.surface)
    /* This message is being delivered to the wrong surface.  */
    return True;

  FinishDndEntry ();

  return True;
}

static Bool
HandleXdndDropEvent (Surface *surface, XEvent *event)
{
  if (event->xclient.data.l[0] != dnd_state.source_window)
    /* The message is coming from the wrong window, or drag and drop
       has not yet been set up.  */
    return True;

  if (surface != dnd_state.surface)
    /* This message is being delivered to the wrong surface.  */
    return True;

  dnd_state.timestamp = event->xclient.data.l[2];
  XLDataDeviceSendDrop (dnd_state.seat, surface);

  /* If finish has already been called, send XdndFinish to the source,
     and complete the transfer.  */
  if (dnd_state.finished
      /* Also respond (but with default values) if the transfer cannot
	 continue because the seat has been destroyed.  */
      || !dnd_state.respond
      || !dnd_state.seat
      /* Also respond if the resource version is less than 3.  */
      || dnd_state.version <= 2)
    RespondToDndDrop ();

  /* Set dnd_state.dropped.  */
  dnd_state.dropped = True;

  return True;
}

void
XLDndWriteAwarenessProperty (Window window)
{
  unsigned long version;

  version = XdndProtocolVersion;
  XChangeProperty (compositor.display, window,
		   XdndAware, XA_ATOM, 32, PropModeReplace,
		   (unsigned char *) &version, 1);
}

/* Keep in mind that the given surface should be a toplevel surface
   with a subcompositor attached.  */

Bool
XLDndFilterClientMessage (Surface *surface, XEvent *event)
{
  if (event->xclient.message_type == XdndEnter)
    return HandleXdndEnterEvent (surface, event);
  else if (event->xclient.message_type == XdndPosition)
    return HandleXdndPositionEvent (surface, event);
  else if (event->xclient.message_type == XdndLeave)
    return HandleXdndLeaveEvent (surface, event);
  else if (event->xclient.message_type == XdndDrop)
    return HandleXdndDropEvent (surface, event);

  return False;
}
